<?php
/**
 * Plugin Name: Bankful
 * Plugin URI: https://bankful.com/integrations-woocommerce/
 * Description: Accept credit card payments, bankful hosted payments page and supports WooCommerce subscriptions on your store.
 * Version: 3.0.0
 * Author: Bankful
 * Author URI: https://bankful.com/
 * Developer: Bankful
 * Developer URI: https://bankful.com/
 * Text Domain: bankful
 *
 * WC requires at least: 7.0
 * WC tested up to: 8.5.2
 *
 * License: GNU General Public License v3.0
 * License URI: http://www.gnu.org/licenses/gpl-3.0.html
 */

if (!defined('ABSPATH')) { // To check absolute path
	exit;
}

function verifyHashSignatureFromRequest( $params ) {
	$bankful_options = get_option('bankful_options');
	$testEnv = ( strtolower($bankful_options['bankful_field_environment']) == strtolower('Test') );
	$req_password   = $testEnv ? ( isset($bankful_options['bankful_field_test_password']) ? $bankful_options['bankful_field_test_password'] : '' ) : ( isset($bankful_options['bankful_field_live_password']) ? $bankful_options['bankful_field_live_password'] : '' );
	if (!$req_password) {
		return false;
	}
	$signatureGenerated = generateSignatureHash($params, $req_password);

	// write_logs_bf('SIGNATURE GENERATED--------------------------');
	// write_logs_bf($signatureGenerated);
	// write_logs_bf($params['SIGNATURE']);
	// write_logs_bf('SIGNATURE--------------------------');
	return $signatureGenerated === $params['SIGNATURE'];
}

function generateSignatureHash( $data, $salt ) {
	$notInclude = array( 'SIGNATURE', 'RECURRING_SCHEDULE_ID' );

	$strField = '';

	$paramData = array_filter(array_keys($data), function ( $key ) use ( $notInclude ) {
		return !in_array($key, $notInclude);
	});

	sort($paramData);

	foreach ($paramData as $paramName) {
		$paramValue = $data[$paramName];
		if (is_array($paramValue) || is_object($paramValue)) {
			$strField .= $paramName . json_encode($paramValue);
		} elseif (is_string($paramValue) || is_numeric($paramValue) || is_bool($paramValue)) {
			if (is_string($paramValue) && '' !== $paramValue) {
				$strField .= $paramName . $paramValue;
			} elseif (!is_string($paramValue)) {
				$strField .= $paramName . $paramValue;
			}
		} elseif ('' !== $paramValue) {
			$strField .= $paramName . $paramValue;
		}
	}

	// write_logs_bf(print_r($paramData, true));
	// foreach ($paramData as $paramName) {
	//  if ('' !== $data[$paramName]) {
	//      $strField .= $paramName . $data[$paramName];
	//  }
	// }
	// MAKE HASH SHA 256
	// write_logs_bf(print_r($strField, true));

	$generatedSignature = hash_hmac('sha256', $strField, $salt);

	return $generatedSignature;
}

// Load all Classes and actions / filters.
add_action('plugins_loaded', 'Bf_Init_Gateway_classes');
add_action('plugins_loaded', 'init_subscription_core_for_bankful', 0);

// Global Variables
$WP_SITE_URL = get_site_url();
$BF_PLUGIN_SOURCE = 'woo_market';
$LIVE_API_SRV = 'https://api.paybybankful.com/';
$SANDBOX_API_SRV = 'https://api-qa.bankfulportal.com/';
// $LOCAL_API_SRV = 'http://localhost:50001/';
// $NGROK_API_SRV = 'https://apibankful.in.ngrok.io/';
// $SANDBOX_API_SRV = $LOCAL_API_SRV;

/**
 * The class itself, please note that it is inside plugins_loaded action hook
 *
 */
function Bf_Init_Gateway_classes() {

	// Check class WP_Payment_Gateway exist or not
	if (!class_exists('WC_Payment_Gateway')) {
		return;
	}

	if (class_exists('WC_Bankful_WS_Gateway') || class_exists('WC_Gateway_Bankful')) {
		add_action('admin_notices', 'bankful_gateway_exists_error');
		return;
	}

	// Credit Card Gateway with Woo Subscriptions
	require_once plugin_dir_path(__FILE__) . 'includes/class-wc-bankful-ws-gateway.php';

	// Hosted Page Bankful Gateway
	require_once plugin_dir_path(__FILE__) . '/includes/bankful-webhook-controller.php';
	require_once plugin_dir_path(__FILE__) . '/includes/class-wc-bankful-hosted-gateway.php';
	require_once plugin_dir_path(__FILE__) . '/includes/blocks/class-wc-bankful-payments-blocks.php';
	require_once plugin_dir_path(__FILE__) . '/includes/blocks/class-wc-bankful-ws-payments-blocks.php';
	// Actions and Filters
	add_action('admin_menu', 'add_bankful_menus', 60);
	add_action('admin_enqueue_scripts', 'enqueue_styles_scripts_bankful');
	add_action('wp_enqueue_scripts', 'front_enqueue_styles_bankful');
	add_action('wp_ajax_sync_subscription_products', 'sync_subscription_products');

	add_action('wp_ajax_migrate_subscription_products', 'migrate_subscription_products');

	add_filter('woocommerce_payment_gateways', 'Bf_Add_Gateway_classes');

	add_action('woocommerce_blocks_loaded', 'woocommerce_gateway_bankful_woocommerce_block_support');
	add_filter('http_request_timeout', 'wp_timeout_extend_fb_one');
	add_filter('plugin_action_links_' . plugin_basename(__FILE__), 'plugin_action_links_bankful');
	
	add_action( 'woocommerce_scheduled_subscription_payment', 'gateway_scheduled_subscription_payment', 10, 1 );
}

if (!function_exists('init_subscription_core_for_bankful')) {
	function init_subscription_core_for_bankful() {
		$options = get_option( 'bankful_options' );
		if ( isset($options['bankful_field_enable_subscription']) && 'enable' === strtolower($options['bankful_field_enable_subscription']) && !class_exists( 'WC_Subscriptions' )) {
			require plugin_dir_path( __FILE__ ) . 'includes/vendor/woocommerce/subscriptions-core/includes/class-wc-subscriptions-core-plugin.php';
			new WC_Subscriptions_Core_Plugin();
		}
	}
}

/**
 * Fire a gateway specific hook for when a subscription payment is due.
 *
 * @since 1.0
 */
function gateway_scheduled_subscription_payment( $subscription_id ) {
	if ( ! is_object( $subscription_id ) ) {
		$subscription = wcs_get_subscription( $subscription_id );
	} else {
		$subscription = $subscription_id;
	}
	if ( false === $subscription ) {
		// translators: %d: subscription ID.
		throw new InvalidArgumentException( sprintf( esc_html( 'Subscription doesn\'t exist in scheduled action: %d', 'woocommerce-subscriptions' ), esc_html($subscription_id) ) );
	}

	if ( ! $subscription->is_manual() && ! $subscription->has_status( wcs_get_subscription_ended_statuses() ) ) {
		trigger_bankful_gateway_renewal_payment_hook( $subscription->get_last_order( 'all', 'renewal' ) );
	}
}

/**
 * Fire a gateway specific hook for when a subscription renewal payment is due.
 *
 * @param WC_Order $renewal_order The renewal order to trigger the payment gateway hook for.
 * @since 2.1.0
 */
function trigger_bankful_gateway_renewal_payment_hook( $renewal_order ) {
	if ( ! empty( $renewal_order ) && $renewal_order->get_total() > 0 && $renewal_order->get_payment_method() === 'bankful_credit_card' ) {

		// Make sure gateways are setup
		WC()->payment_gateways();

		/**
		 * Do Bankful Specific Scheduled Payment Action Hook
		 * 
		 * @since 3.0.0
		 * @param string Order total amount 
		 * @param WC_Order $renewal_order  The renewal order to trigger the payment gateway hook for.
		 */
		do_action( 'woocommerce_scheduled_subscription_payment_bankful_credit_card', $renewal_order->get_total(), $renewal_order );
	} else if ( ! empty( $renewal_order ) && $renewal_order->get_total() > 0 && $renewal_order->get_payment_method() === 'bankful_hosted_gateway' ) {

		// Make sure gateways are setup
		WC()->payment_gateways();

		/**
		 * Do Bankful Specific Scheduled Payment Action Hook
		 * 
		 * @since 3.0.0
		 * @param string Order total amount 
		 * @param WC_Order $renewal_order  The renewal order to trigger the payment gateway hook for.
		 */
		do_action( 'woocommerce_scheduled_subscription_payment_bankful_hosted_gateway', $renewal_order->get_total(), $renewal_order );
	}
}

/**
 * Plugin url.
 *
 * @return string
 */
function plugin_url() {
	return untrailingslashit(plugins_url('/', __FILE__));
}

/**
 * Plugin url.
 *
 * @return string
 */
function plugin_abspath() {
	return trailingslashit(plugin_dir_path(__FILE__));
}

/**
 * Registers WooCommerce Blocks integration.
 *
 */
function woocommerce_gateway_bankful_woocommerce_block_support() {
	$options = get_option('bankful_options');

	if (isset($options['bankful_field_environment']) && 'Test' == $options['bankful_field_environment']) {
		if (isset($options['bankful_field_test_validated_account_cc']) && 'yes' == $options['bankful_field_test_validated_account_cc']) {
			if (class_exists('Automattic\WooCommerce\Blocks\Payments\Integrations\AbstractPaymentMethodType')) {
				add_action(
					'woocommerce_blocks_payment_method_type_registration',
					function ( Automattic\WooCommerce\Blocks\Payments\PaymentMethodRegistry $payment_method_registry ) {
						$payment_method_registry->register(new WC_Bankful_WS_Gateway_Blocks_Support());
					}
				);
			}
		}
		if (isset($options['bankful_field_test_validated_account_hg']) && 'yes' == $options['bankful_field_test_validated_account_hg']) {
			if (class_exists('Automattic\WooCommerce\Blocks\Payments\Integrations\AbstractPaymentMethodType')) {
				add_action(
					'woocommerce_blocks_payment_method_type_registration',
					function ( Automattic\WooCommerce\Blocks\Payments\PaymentMethodRegistry $payment_method_registry ) {
						$payment_method_registry->register(new WC_Gateway_Bankful_Blocks_Support());
					}
				);
			}
		}
	} else if (isset($options['bankful_field_environment']) && 'Live' == $options['bankful_field_environment']) {

		if (isset($options['bankful_field_live_validated_account_cc']) && 'yes' == $options['bankful_field_live_validated_account_cc']) {
			if (class_exists('Automattic\WooCommerce\Blocks\Payments\Integrations\AbstractPaymentMethodType')) {
				add_action(
					'woocommerce_blocks_payment_method_type_registration',
					function ( Automattic\WooCommerce\Blocks\Payments\PaymentMethodRegistry $payment_method_registry ) {
						$payment_method_registry->register(new WC_Bankful_WS_Gateway_Blocks_Support());
					}
				);
			}
		}
		if (isset($options['bankful_field_live_validated_account_hg']) && 'yes' == $options['bankful_field_live_validated_account_hg']) {
			if (class_exists('Automattic\WooCommerce\Blocks\Payments\Integrations\AbstractPaymentMethodType')) {
				add_action(
					'woocommerce_blocks_payment_method_type_registration',
					function ( Automattic\WooCommerce\Blocks\Payments\PaymentMethodRegistry $payment_method_registry ) {
						$payment_method_registry->register(new WC_Gateway_Bankful_Blocks_Support());
					}
				);
			}
		}
	}
}

/**
 * Enqueue custom CSS and scripts for Bankful Gateway Frontend
 *
 */
function front_enqueue_styles_bankful() {
	wp_register_style('bankful_front_screens', plugin_dir_url(__FILE__) . 'assets/css/bf-style.css', false, '3.0.0');
	wp_enqueue_style('bankful_front_screens');
}


/**
 * Add links to plugins page for checkout settings
 *
 * @param  array $links
 * @return array
 */
function plugin_action_links_bankful( $links = array() ) {

	$options = get_option('bankful_options');
	$plugin_links = array( '<a href="' . admin_url('admin.php?page=bankful_settings') . '">' . __('Bankful', 'bankful') . '</a>' );

	if (isset($options['bankful_field_environment']) && 'Test' == $options['bankful_field_environment']) {

		( isset($options['bankful_field_test_validated_account_cc']) && 'yes' == $options['bankful_field_test_validated_account_cc'] ) ?
			$plugin_links[] = '<a href="' . admin_url('admin.php?page=wc-settings&tab=checkout&section=bankful_credit_card') . '">' . __('Credit Card', 'bankful') . '</a>' : '';
		( isset($options['bankful_field_test_validated_account_hg']) && 'yes' == $options['bankful_field_test_validated_account_hg'] ) ?
			$plugin_links[] = '<a href="' . admin_url('admin.php?page=wc-settings&tab=checkout&section=bankful_hosted_gateway') . '">' . __('Bankful Hosted', 'bankful') . '</a>' : '';
	} else if (isset($options['bankful_field_environment']) && 'Live' == $options['bankful_field_environment']) {

		( isset($options['bankful_field_live_validated_account_cc']) && 'yes' == $options['bankful_field_live_validated_account_cc'] ) ?
			$plugin_links[] = '<a href="' . admin_url('admin.php?page=wc-settings&tab=checkout&section=bankful_credit_card') . '">' . __('Credit Card', 'bankful') . '</a>' : '';
		( isset($options['bankful_field_live_validated_account_hg']) && 'yes' == $options['bankful_field_live_validated_account_hg'] ) ?
			$plugin_links[] = '<a href="' . admin_url('admin.php?page=wc-settings&tab=checkout&section=bankful_hosted_gateway') . '">' . __('Bankful Hosted', 'bankful') . '</a>' : '';
	}

	return array_merge($plugin_links, $links);
}

/**
 * This function is used to log on globally for bankful plugin
 *
 * @param string $message message to log
 *
 * @param string $level log level like info/debug
 */
function write_logs_bf( $message = '', $level = 'info' ) {
	$logger = wc_get_logger();

	$logger->log($level, wc_print_r($message, true), array( 'source' => 'bankful_payment_gateway' ));
}

/**
 * This action hook registers notice that bankful gateway plugin already active
 *
 * @param string $gateways SDK/API method name
 *
 * @return string error message
 */
function bankful_gateway_exists_error() {
	$return_error = '<div class="notice error my-acf-notice is-dismissible" >';
	$return_error .= '<p>' . esc_html_e('Another Bankful Gateway already activated, please deactivate other Bankful or Bankful Hosted Gateway. Thanks.', 'bankful') . '</p>';
	$return_error .= '</div>';
	return $return_error;
}

/**
 * This action hook registers our PHP class as a WooCommerce payment gateway
 *
 * @param string $gateways SDK/API method name
 *
 * @return object  $this
 */
function Bf_Add_Gateway_classes( $gateways ) {
	$options = get_option('bankful_options');

	if (isset($options['bankful_field_environment']) && 'Test' == $options['bankful_field_environment']) {

		( isset($options['bankful_field_test_validated_account_cc']) && 'yes' == $options['bankful_field_test_validated_account_cc'] ) ?
			$gateways[] = 'WC_Bankful_WS_Gateway' : '';
		( isset($options['bankful_field_test_validated_account_hg']) && 'yes' == $options['bankful_field_test_validated_account_hg'] ) ?
			$gateways[] = 'WC_Gateway_Bankful' : '';
	} else if (isset($options['bankful_field_environment']) && 'Live' == $options['bankful_field_environment']) {

		( isset($options['bankful_field_live_validated_account_cc']) && 'yes' == $options['bankful_field_live_validated_account_cc'] ) ?
			$gateways[] = 'WC_Bankful_WS_Gateway' : '';
		( isset($options['bankful_field_live_validated_account_hg']) && 'yes' == $options['bankful_field_live_validated_account_hg'] ) ?
			$gateways[] = 'WC_Gateway_Bankful' : '';
	}

	return $gateways;
}

/**
 * Process the payment and return the result.
 *
 * @param datetime $time Order ID.
 *
 * @return array
 */
function wp_timeout_extend_fb_one( $time ) {
	// Default timeout is 5
	return 30;
}

/**
 * Add Bankful Menus to Admin Menus
 *
 */
function add_bankful_menus() {
	add_submenu_page(
		'woocommerce',
		'Bankful',
		'Bankful',
		'manage_woocommerce',
		'bankful_settings',
		'bankful_options_page_html'
	);
}


/**
 * Top level menu callback function for bankful_options_page_html
 */
function bankful_options_page_html() {
	global $SANDBOX_API_SRV;
	global $LIVE_API_SRV;
	// check user capabilities
	if (!current_user_can('manage_options')) {
		return;
	}

	// check if the user have submitted the settings
	// WordPress will add the "settings-updated" $_GET parameter to the url
	if (isset($_GET['settings-updated'])) {
		// add settings saved message with the class of "updated"
		add_settings_error('bankful_messages', 'bankful_message', __('Settings Saved', 'bankful'), 'updated');
	}

	$options = get_option('bankful_options');

	// show error/update messages
	settings_errors('bankful_messages');
	?>
	<div class="wrap">
		<h1><?php echo esc_html(get_admin_page_title()); ?></h1>
		<form action="options.php" method="post" name="bankful_settings">
			<?php
			// output security fields for the registered setting "bankful"
			settings_fields('bankful');
			// output setting sections and their fields
			// (sections are registered for "bankful", each field is registered to a specific section)
			do_settings_sections('bankful');
			?>
			<input type="button" name="submitValidate" id="submitValidate" class="button" value="Sync Settings">
			<p>
				<?php
				if (isset($options['bankful_field_environment']) && 'Test' == $options['bankful_field_environment']) {

					echo ( ( isset($options['bankful_field_test_validated_account_cc']) && 'yes' == $options['bankful_field_test_validated_account_cc'] ) ?
						'<a style="margin-right: 10px;" type="button" name="redirectToCC" id="redirectToCC" class="button" href="../wp-admin/admin.php?page=wc-settings&tab=checkout&section=bankful_credit_card">Credit Card Settings</a>' : '' );
					echo ( ( isset($options['bankful_field_test_validated_account_hg']) && 'yes' == $options['bankful_field_test_validated_account_hg'] ) ?
						'<a type="button" name="redirectToHP" id="redirectToHP" class="button" href="../wp-admin/admin.php?page=wc-settings&tab=checkout&section=bankful_hosted_gateway">Hosted Page Settings</a>' : '' );
				} else if (isset($options['bankful_field_environment']) && 'Live' == $options['bankful_field_environment']) {

				echo( ( isset($options['bankful_field_live_validated_account_cc']) && 'yes' == $options['bankful_field_live_validated_account_cc'] ) ?
				'<a style="margin-right: 10px;" type="button" name="redirectToCC" id="redirectToCC" class="button" href="">Credit Card Settings</a>' : '' );
				echo( ( isset($options['bankful_field_live_validated_account_hg']) && 'yes' == $options['bankful_field_live_validated_account_hg'] ) ?
				'<a type="button" name="redirectToHP" id="redirectToHP" class="button" href="">Hosted Page Settings</a>' : '' );
				}
			echo '</p><p><input type="button" name="syncSubsProducts" id="syncSubsProducts" class="button" value="Sync Subscription Products"/></p><p>';
			echo '</p><p><input type="button" name="migrateSubsProducts" id="migrateSubsProducts" class="button" value="Migrate Recurring Products"/></p><p>';
			// output save settings button
			submit_button( 'Save Settings', 'hidden' );
				?>
			</p>
		</form>
	</div>
	<script>
		jQuery('#syncSubsProducts').on('click', function(event) {
			jQuery.ajax({
				type: 'post',
				url: ajaxurl,
				data : {action: "sync_subscription_products"},
				success: function(response) {
					console.log(response);
					const data = response['data'];
					if(data){
						alert("Products Updated: " + data['updatedProducts'] + '. New Created Products: ' + data['createdProducts']);
					}else{
						alert("Request Failed");
					}
				}
			});
		});

		jQuery('#migrateSubsProducts').on('click', function(event) {
			jQuery.ajax({
				type: 'post',
				url: ajaxurl,
				data : {action: "migrate_subscription_products"},
				success: function(response) {
					console.log(response);
					const data = response['updatedCount'];
					alert("Products Updated: " + data + '.');
				}
			});
		});


		jQuery('#submitValidate').on('click', async function(event) {
			const test_username = jQuery("#bankful_field_test_username").val();
			const test_password = jQuery("#bankful_field_test_password").val();

			const live_username = jQuery("#bankful_field_live_username").val();
			const live_password = jQuery("#bankful_field_live_password").val();

			const validated_test_username = `${test_username.substr(0, 2)}*****${test_username.substr(test_username.length - 8)}`;

			const validated_live_username = `${live_username.substr(0, 2)}*****${live_username.substr(live_username.length - 8)}`;

			const environment = jQuery("#bankful_field_environment").val();

			const current_site_url = '<?php echo esc_url(get_site_url()); ?>';

			const request_url_live = '<?php echo esc_url($LIVE_API_SRV); ?>';
			const request_url_test = '<?php echo esc_url($SANDBOX_API_SRV); ?>';

			// console.log(environment);
			// console.log(current_site_url);
			// console.log(request_url_live);
			// console.log(request_url_test);

			let live_saved = false;
			let test_saved = false;
			if (live_username && live_password) {
				await jQuery.ajax({
					type: "post",
					dataType: "json",
					url: request_url_live + 'api/woocommerce/validate-credentials',
					data: {
						req_username: live_username,
						req_password: live_password,
						site_url: current_site_url
					},
					success: function(response) {

						if ('success' === response.status) {
							const data = response.data;
							jQuery("#bankful_field_live_validate").val(validated_live_username);
							jQuery("#bankful_field_live_validated_account_hg").val(data.isHostedPageEnabled === "Y" ? 'yes' : "no");
							jQuery("#bankful_field_live_validated_account_cc").val(data.isNativeEnabled === "Y" ? 'yes' : "no");
							jQuery("#bankful_field_live_subscription_status").val(data.isSubscriptionEnabled === "Y" ? 'yes' : "no");
							jQuery(".description.bankful_field_live_password").html('<span style="color:green;">Validation Success.</span>');

							if (environment === "Live") {
								jQuery('#submit').trigger("click");
							}
						} else {
							jQuery(".description.bankful_field_live_password").html('<span style="color:red;">Validation Failed.</span>');
							jQuery("#bankful_field_live_validate").val("");
							jQuery("#bankful_field_live_validated_account_hg").val("no");
							jQuery("#bankful_field_live_validated_account_cc").val("no");
							jQuery("#bankful_field_live_subscription_status").val("no");
							jQuery("#bankful_field_enable_subscription").val("Disable");
							alert(response.errorMessage);
							jQuery('#submit').trigger("click");
						}
					},
					error: function (jqXHR, exception) {
						var msg = '';
						if (jqXHR.status === 0) {
							msg = 'Not connect.\n Verify Network.';
						} else if (jqXHR.status == 404) {
							msg = 'Requested page not found. [404]';
						} else if (jqXHR.status == 500) {
							msg = 'Internal Server Error [500].';
						} else if (exception === 'parsererror') {
							msg = 'Requested JSON parse failed.';
						} else if (exception === 'timeout') {
							msg = 'Time out error.';
						} else if (exception === 'abort') {
							msg = 'Ajax request aborted.';
						} else {
							responseJSON = JSON.parse(jqXHR.responseText);
							msg = 'Validation Error.\n' + responseJSON.errorMessage;
						}
						jQuery(".description.bankful_field_live_password").html('<span style="color:red;">Validation Failed.</span>');
						jQuery("#bankful_field_live_validate").val("");
						jQuery("#bankful_field_live_validated_account_hg").val("no");
						jQuery("#bankful_field_live_validated_account_cc").val("no");
						jQuery("#bankful_field_live_subscription_status").val("no");
						jQuery("#bankful_field_enable_subscription").val("Disable");
						alert(msg);
						jQuery('#submit').trigger("click");
					}
				});

			} else {
				jQuery(".description.bankful_field_live_password").html('<span style="color:red;">Please enter Live Username and Password.</span>');
			}

			if (test_username && test_password) {
				await jQuery.ajax({
					type: "post",
					dataType: "json",
					url: request_url_test + 'api/woocommerce/validate-credentials',
					data: {
						req_username: test_username,
						req_password: test_password,
						site_url: current_site_url
					},
					success: function(response) {
						// if(response.type == "success") {

						if ('success' === response.status) {
							const data = response.data;
							jQuery("#bankful_field_test_validate").val(validated_test_username);
							jQuery("#bankful_field_test_validated_account_hg").val(data.isHostedPageEnabled === "Y" ? 'yes' : "no");
							jQuery("#bankful_field_test_validated_account_cc").val(data.isNativeEnabled === "Y" ? 'yes' : "no");
							jQuery("#bankful_field_test_subscription_status").val(data.isSubscriptionEnabled === "Y" ? 'yes' : "no");
							jQuery(".description.bankful_field_test_password").html('<span style="color:green;">Validation Success.</span>');

							if (environment === "Test") {
								jQuery('#submit').trigger("click");
							}
						} else {
							jQuery(".description.bankful_field_test_password").html('<span style="color:red;">Validation Failed.</span>');
							jQuery("#bankful_field_test_validate").val("");
							jQuery("#bankful_field_test_validated_account_hg").val("no");
							jQuery("#bankful_field_test_validated_account_cc").val("no");
							jQuery("#bankful_field_test_subscription_status").val("no");
							jQuery("#bankful_field_enable_subscription").val("Disable");
							alert(response.errorMessage);
							jQuery('#submit').trigger("click");
						}
					}
				});

			} else {
				jQuery(".description.bankful_field_live_password").html('<span style="color:red;">Please enter Test Username and Password.</span>');
			}

		});

		jQuery('#TEST_edit').on('click', function(event) {
			event.preventDefault();
			jQuery("#bankful_field_test_validate").val("");
			jQuery("#bankful_field_test_validated_account_hg").val("no");
			jQuery("#bankful_field_test_validated_account_cc").val("no");
			jQuery("#bankful_field_test_subscription_status").val("no");
			jQuery("#bankful_field_enable_subscription").val("Disable");
			jQuery('#submit').trigger("click");
		});

		jQuery('#LIVE_edit').on('click', function(event) {
			event.preventDefault();
			jQuery("#bankful_field_live_validate").val("");
			jQuery("#bankful_field_live_validated_account_hg").val("no");
			jQuery("#bankful_field_live_validated_account_cc").val("no");
			jQuery("#bankful_field_live_subscription_status").val("no");
			jQuery("#bankful_field_enable_subscription").val("Disable");
			jQuery('#submit').trigger("click");
		});
	</script>
<?php
}

/**
 * Redirect to WooCommerce Payment Gateway Settings Page
 *
 */
function redirect_to_gateway_bankful_cc() {
	wp_safe_redirect(admin_url('admin.php?page=wc-settings&tab=checkout&section=bankful_credit_card'));
	exit;
}

/**
 * Redirect to WooCommerce Payment Gateway Settings Page
 *
 */
function redirect_to_gateway_bankful_hg() {
	wp_safe_redirect(admin_url('admin.php?page=wc-settings&tab=checkout&section=bankful_hosted_gateway'));
	exit;
}

/**
 * Enqueue custom CSS and scripts for Bankful Gateeway admin
 *
 */
function enqueue_styles_scripts_bankful() {
	wp_register_style('bankful_admin_screens', plugin_dir_url(__FILE__) . 'assets/css/admin-bf.css', false, '3.0.0');
	wp_enqueue_style('bankful_admin_screens');
}

/**
 * Register our bankful_settings_init to the admin_init action hook.
 */
add_action('admin_init', 'bankful_settings_init');

/**
 * Bankful option and settings
 */
function bankful_settings_init() {
	// Register a new setting for "bankful" page.
	register_setting('bankful', 'bankful_options');

	$options = get_option('bankful_options');
	$hidden_validated = ( isset($options['bankful_field_test_validate']) && !empty($options['bankful_field_test_validate']) ) ? 'hidden' : '';
	$hidden_not_validated = ( !isset($options['bankful_field_test_validate']) || empty($options['bankful_field_test_validate']) ) ? 'hidden' : '';

	$hidden_validated_live = ( isset($options['bankful_field_live_validate']) && !empty($options['bankful_field_live_validate']) ) ? 'hidden' : '';
	$hidden_not_validated_live = ( !isset($options['bankful_field_live_validate']) || empty($options['bankful_field_live_validate']) ) ? 'hidden' : '';

	$hidden_subscription = ( ( isset($options['bankful_field_test_subscription_status']) && 'yes' === $options['bankful_field_test_subscription_status'] ) ||
		( isset($options['bankful_field_live_subscription_status']) && 'yes' === $options['bankful_field_live_subscription_status'] ) ) ? '' : 'hidden';

	// Register a new section in the "bankful" page.
	add_settings_section(
		'bankful_section_settings',
		__('API Credentials', 'bankful'),
		'bankful_section_settings_callback',
		'bankful'
	);

	// Register a new field in the "bankful_field_test_cb" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_environment',
		__('Current Environment:', 'bankful'),
		'bankful_field_dropdown',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_environment',
			'class'             => 'bankful_row',
			'type'              => 'text',
			'desc'              => '',
			'default'           => 'Live',
		)
	);

	// Register a new field in the "bankful_field_test_cb" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_test_validate',
		__('Test API Status:', 'bankful'),
		'bankful_field_test_cb',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_test_validate',
			'class'             => 'bankful_row disabled ' . $hidden_not_validated,
			'type'              => 'text',
			'desc'              => '',
			'default'           => '',
			'info'              => 'TEST',
		)
	);

	// Register a new field in the "bankful_field_test_cb" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_test_validated_account_cc',
		__('Test API Connected Account Credit Card:', 'bankful'),
		'bankful_field_test_cb',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_test_validated_account_cc',
			'class'             => 'bankful_row hidden ',
			'type'              => 'text',
			'desc'              => '',
			'default'           => 'no',
		)
	);

	// Register a new field in the "bankful_field_test_cb" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_test_validated_account_hg',
		__('Test API Connected Account Hosted Gateway:', 'bankful'),
		'bankful_field_test_cb',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_test_validated_account_hg',
			'class'             => 'bankful_row hidden ',
			'type'              => 'text',
			'desc'              => '',
			'default'           => 'no',
		)
	);

	// Register a new field in the "bankful_field_test_cb" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_test_username',
		__('Test API Username', 'bankful'),
		'bankful_field_test_cb',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_test_username',
			'class'             => 'bankful_row ' . $hidden_validated,
			'type' => 'text',
			'desc' => 'Add Test Username here.',
			'default' => 'sandbox-test@bankful.com',
		)
	);

	// Register a new field in the "bankful_field_test_cb" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_test_password',
		__('Test API Password', 'bankful'),
		'bankful_field_test_cb',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_test_password',
			'class'             => 'bankful_row ' . $hidden_validated,
			'type' => 'password',
			'desc' => 'Add Test Password here.',
			'default' => 'tuMcC7m333J4VMsb',
		)
	);

	// Register a new field in the "bankful_field_test_cb" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_live_validate',
		__('Live API Status:', 'bankful'),
		'bankful_field_test_cb',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_live_validate',
			'class'             => 'bankful_row disabled ' . $hidden_not_validated_live,
			'type'              => 'text',
			'desc'              => '',
			'default'           => '',
			'info'              => 'LIVE',
		)
	);

	// Register a new field in the "bankful_field_test_cb" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_live_validated_account_cc',
		__('Live API Connected Account Credit Card:', 'bankful'),
		'bankful_field_test_cb',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_live_validated_account_cc',
			'class'             => 'bankful_row hidden ',
			'type'              => 'text',
			'desc'              => '',
			'default'           => 'no',
		)
	);

	// Register a new field in the "bankful_field_test_cb" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_live_validated_account_hg',
		__('Live API Connected Account Hosted Gateway:', 'bankful'),
		'bankful_field_test_cb',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_live_validated_account_hg',
			'class'             => 'bankful_row hidden ',
			'type'              => 'text',
			'desc'              => '',
			'default'           => 'no',
		)
	);

	// Register a new field in the "bankful_field_test_cb" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_live_username',
		__('LIVE API Username', 'bankful'),
		'bankful_field_test_cb',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_live_username',
			'class'             => 'bankful_row ' . $hidden_validated_live,
			'type' => 'text',
			'desc' => 'Add Live Username here.',
			'default' => '',
		)
	);

	// Register a new field in the "bankful_field_test_cb" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_live_password',
		__('LIVE API Password', 'bankful'),
		'bankful_field_test_cb',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_live_password',
			'class'             => 'bankful_row ' . $hidden_validated_live,
			'type' => 'password',
			'desc' => 'Add Live Password here.',
			'default' => '',
		)
	);

	// Register a new field in the "bankful_field_live_subscription_status" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_live_subscription_status',
		__('Live Subscription Status', 'bankful'),
		'bankful_field_test_cb',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_live_subscription_status',
			'class'             => 'bankful_row hidden ',
			'type'              => 'text',
			'desc'              => '',
			'default'           => 'no',
		)
	);

	// Register a new field in the "bankful_field_test_subscription_status" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_test_subscription_status',
		__('Test Subscription Status', 'bankful'),
		'bankful_field_test_cb',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_test_subscription_status',
			'class'             => 'bankful_row hidden ',
			'type'              => 'text',
			'desc'              => '',
			'default'           => 'no',
		)
	);

	// Register a new field in the "bankful_field_test_cb" section, inside the "bankful" page.
	add_settings_field(
		'bankful_field_enable_subscription',
		__( 'Bankful Subscription Module:', 'bankful' ),
		'bankful_field_dropdown_subscription',
		'bankful',
		'bankful_section_settings',
		array(
			'label_for'         => 'bankful_field_enable_subscription',
			'class'             => 'bankful_row ' . $hidden_subscription,
			'type'              => 'text',
			'desc'              => 'Bankful Subscription Module for Bankful Payment Gateway',
			'default'           => 'Disable',
		)
	);
};

/**
 * Bankful Select Environment field callback function.
 *
 * @param array $args
 */
function bankful_field_dropdown( $args ) {
	$options = get_option('bankful_options');
	$items = array( 'Live', 'Test' );
	echo "<select id='bankful_field_environment' name='bankful_options[bankful_field_environment]'>";
	foreach ($items as $item) {
		$selected = ( strtolower($options['bankful_field_environment']) == strtolower($item) ) ? 'selected="selected"' : '';
		echo "<option value='" . esc_attr($item) . "' " . esc_attr($selected) . '>' . esc_html($item) . '</option>';
	}
	?>
	</select>
<?php
}

/**
 * Bankful Select Environment field callback function.
 *
 * @param array $args
 */
function bankful_field_dropdown_subscription( $args ) {
	$options = get_option('bankful_options');
	$items = array( 'Enable', 'Disable' );
	$wc_sub = false;
	// $wc_sub_all_products = false;
	if ( is_plugin_active( 'woocommerce-subscriptions/woocommerce-subscriptions.php' ) ) {
		$wc_sub = true;
	} 
	// if ( is_plugin_active( 'woocommerce-all-products-for-subscriptions/woocommerce-all-products-for-subscriptions.php' ) ) {
	//  $wc_sub_all_products = true;
	// } 
	echo "<select id='bankful_field_enable_subscription' name='bankful_options[bankful_field_enable_subscription]' " . ( $wc_sub ? 'disabled' : '' ) . '>';
	foreach ($items as $item) {
		$selected = ( strtolower($options['bankful_field_enable_subscription'])==strtolower($item) ) ? 'selected="selected"' : '';
		echo "<option value='" . esc_attr($item) . "' " . esc_attr($selected) . '>' . esc_html($item) . '</option>';
	}
	?>
	</select>
	<?php
	if ( $wc_sub ) {
		echo '<span>Please Disable WooCommerce Subscription Plugin to work with Bankful Subscriptions</span>';
	} 
	// if ( $wc_sub_all_products ) {
	//  echo '<span>Please Disable WooCommerce All Products Subscription Plugin to work with Bankful Subscriptions</span>';
	// } 
}

// add_filter( 'woocommerce_add_to_cart_validation', 'sold_individually_for_reccuring_product_bf', 10, 3 );
// function sold_individually_for_reccuring_product_bf( $passed, $product_id, $quantity) {
//  write_logs_bf('sold_individually_for_reccuring_product_bf');
//  write_logs_bf(WC()->cart->get_cart_contents_count());
//  $passed = true;
//  $product = wc_get_product($product_id);
//  if ( ( WC()->cart->get_cart_contents_count() > 0 && class_exists( 'WC_Subscriptions_Product' ) && WC_Subscriptions_Product::is_subscription( $product_id ) ) || ( WC()->cart->get_cart_contents_count() > 0 && 'simple' === $product->get_type() && class_exists( 'WCS_ATT_Product_Schemes' ) && WCS_ATT_Product_Schemes::get_subscription_schemes( $product ) )) {
//      $passed = false;
//      $passed = check_recurring_in_cart_wc($product_id, $product);
//  } else {
//      $passed = check_recurring_in_cart_wc($product_id, $product);
//  }

//  if ( !$passed ) {
//      $message = __('The Subscription Product Can Only be sold individually and cart is not empty.', 'woocommerce');
//      wc_add_notice( $message, 'error' );
//  }
//  return $passed;
// }

// function check_recurring_in_cart_wc( $product_id, $product ) {
//  $has_recurring = true;

//  foreach ( WC()->cart->get_cart() as $item ) {
//      $is_recurring = ( class_exists( 'WC_Subscriptions_Product' ) && WC_Subscriptions_Product::is_subscription( $item['product_id'] ) ) || ( 'simple' === $product->get_type() && class_exists( 'WCS_ATT_Product_Schemes' ) && WCS_ATT_Product_Schemes::get_subscription_schemes( $product ) );

//      if ( $is_recurring ) {
//          $has_recurring = false;
//      }
//      // Allowing the same product to be added (not activated. you can uncomment lines below)
//      if ( $item['product_id'] == $product_id  ) {
//          $has_recurring = true;
//      }

//      if ( $has_recurring ) {
//          break; // stops the 1st loop
//      }
//  }
//  return $has_recurring;
// }

// add_filter( 'wcsatt_show_cart_item_options', '__return_false' );


function sync_subscription_products() {
	global $LIVE_API_SRV;
	global $SANDBOX_API_SRV;

	$options = get_option( 'bankful_options' );
	$site_id = get_option( 'bankful_options_site_id' );
	$environment = $options['bankful_field_environment'];
	$currency_store = get_option('woocommerce_currency');
	
	$api_endpoint = strtolower($environment) === 'test' ? $SANDBOX_API_SRV . 'api/' : $LIVE_API_SRV . 'api/';
	$request_url = $api_endpoint . 'woocommerce/sync-products';
	$request_site_id = $api_endpoint . 'woocommerce/register-site';
	$req_username = strtolower($environment) === 'test' ? $options['bankful_field_test_username'] : $options['bankful_field_live_username'];
	$req_password = strtolower($environment) === 'test' ? $options['bankful_field_test_password'] : $options['bankful_field_live_password'];

	// $bankful_site_id = $options['bankful_site_id'];
	if (!isset($site_id) || empty($site_id)) {
		$response = wp_remote_post(
			$request_site_id,
			array(
				'body' => array( 'req_username' => $req_username, 'req_password' => $req_password, 'site_url' => get_site_url(), 'site_title' => get_bloginfo('name') ),
				'headers'     => array(
					'Content-Type' => 'application/x-www-form-urlencoded',
				),
			)
		);
		$body = wp_remote_retrieve_body($response);
		$json_response = json_decode($body);
		$data_response = $json_response->data;
		if (isset($data_response) && !empty($data_response)) {
			write_logs_bf($data_response, 'info');
			if ($data_response->site_id) {
				write_logs_bf($data_response->site_id, 'info');
				add_option('bankful_options_site_id', $data_response->site_id);
				update_option('bankful_options_site_id', $data_response->site_id);
			}
		}
		write_logs_bf($body, 'info');
	}

	if (isset($site_id) && !empty($site_id)) {
		$products = wc_get_products( array( 'type' => array( 'simple', 'subscription', 'subscription_variation', 'variable-subscription' ), 'status' => 'publish', 'limit' => -1 ) );

		$products_array = array();
		foreach ($products as $product) {
			write_logs_bf('================================');
			write_logs_bf($product->get_title());
			if (class_exists( 'WCS_ATT_Product_Schemes' ) && WCS_ATT_Product_Schemes::get_subscription_schemes( $product )) {
				$subscriptionSchemes = WCS_ATT_Product_Schemes::get_subscription_schemes( $product );
			}
			if ('simple' === $product->get_type() && !empty($subscriptionSchemes)) {
				foreach ($subscriptionSchemes as $subscriptionScheme) {
					$schemeData = $subscriptionScheme->get_data();
					$products_array[] = array(
						'id' => $product->get_ID() . '_' . $subscriptionScheme->get_key(),
						'title' => $product->get_title() . '_' . $subscriptionScheme->get_key(),
						'price' => $schemeData['price'],
						'type' => $product->get_type(),
						'image' => wp_get_attachment_url($product->get_image_id()),
						'description' => $product->get_description(),
						'date_created' => $product->get_date_created()->date('Y-m-d H:i:s'),
						'interval' => $subscriptionScheme->get_interval(),
						'period' => $subscriptionScheme->get_period(),
						'length' => $subscriptionScheme->get_length(),
					); 
					// write_logs_bf($subscriptionScheme->get_key());
					// write_logs_bf($subscriptionScheme->get_data()['price']);
					// write_logs_bf($subscriptionScheme->get_interval());
					// write_logs_bf($subscriptionScheme->get_length());
					// write_logs_bf($subscriptionScheme->get_sale_price());
					// write_logs_bf($subscriptionScheme->get_period());
					// write_logs_bf($subscriptionScheme->get_discount());
					// write_logs_bf($subscriptionScheme->get_regular_price());
				}
			} else if ('simple' !== $product->get_type()) {
				// $products_array[] = array(
				//  'id' => $product->get_ID(),
				//  'title' => $product->get_title(),
				//  'price' => $product->get_price(),
				//  'type' => $product->get_type(),
				//  'image' => wp_get_attachment_url($product->get_image_id()),
				//  'description' => $product->get_description(),
				//  'date_created' => $product->get_date_created()->date('Y-m-d H:i:s'),
				//  'interval' => WC_Subscriptions_Product::get_interval( $product ),
				//  'period' => WC_Subscriptions_Product::get_period( $product ),
				//  'length' => WC_Subscriptions_Product::get_length( $product )
				// );
				write_logs_bf($product->get_type());

				if ($product->is_type('subscription')) {
					$products_array[] = array(
						'id' => $product->get_ID(),
						'title' => $product->get_title(),
						'price' => $product->get_price(),
						'type' => $product->get_type(),
						'image' => wp_get_attachment_url($product->get_image_id()),
						'description' => $product->get_description(),
						'date_created' => $product->get_date_created()->date('Y-m-d H:i:s'),
						'interval' => WC_Subscriptions_Product::get_interval($product),
						'period' => WC_Subscriptions_Product::get_period($product),
						'length' => WC_Subscriptions_Product::get_length($product),
					);
					write_logs_bf($product->get_type());
				} elseif ($product->is_type('variable-subscription')) {
					$variation_ids = $product->get_children();
					write_logs_bf(print_r($variation_ids, true));
					foreach ($variation_ids as $variation_id) {
						$variation = wc_get_product($variation_id);

						$attributes = $variation->get_attributes();
						$attributes_string = '';
						foreach ($attributes as $key => $value) {
							$term = get_term_by('slug', $value, $key);
							if ($term) {
								$attributes_string .= ' - ' . $term->name;
							} else {
								$attributes_string .= ' - ' . ucfirst($value);
							}
						}
			
						$products_array[] = array(
							'id' => $variation->get_ID(),
							'title' => $variation->get_title() . $attributes_string,
							'price' => $variation->get_price(),
							'type' => $variation->get_type(),
							'image' => wp_get_attachment_url($variation->get_image_id()),
							'description' => $variation->get_description(),
							'date_created' => $variation->get_date_created()->date('Y-m-d H:i:s'),
							'interval' => WC_Subscriptions_Product::get_interval($variation),
							'period' => WC_Subscriptions_Product::get_period($variation),
							'length' => WC_Subscriptions_Product::get_length($variation),
						);
					}
				}

			}
		}

		write_logs_bf(print_r($products_array, true));
		// write_logs_bf($request_url);

		$response = wp_remote_post(
			$request_url,
			array(
				'body' => array( 'products' => $products_array, 'currency' => $currency_store, 'site_id' => $site_id, 'req_username' => $req_username, 'req_password' => $req_password ),
				'headers'     => array(
					'Content-Type' => 'application/x-www-form-urlencoded',
				),
			)
		);

		// write_logs_bf(print_r($response, true));

		$body = wp_remote_retrieve_body($response);
		$json_response = json_decode($body);
		// write_logs_bf(print_r($json_response, true));
		wp_send_json( $json_response, 200, null );
	}
	return null;
}

function migrate_subscription_products() {
	global $LIVE_API_SRV;
	global $SANDBOX_API_SRV;
	if ( ! class_exists( 'WC_Subscriptions_Product' ) ) {
		return false;
	}
	
	$options = get_option( 'bankful_options' );
	$site_id = get_option( 'bankful_options_site_id' );
	$environment = $options['bankful_field_environment'];
	
	$api_endpoint = strtolower($environment) === 'test' ? $SANDBOX_API_SRV . 'api/' : $LIVE_API_SRV . 'api/';

	$request_site_id = $api_endpoint . 'woocommerce/register-site';
	$req_username = strtolower($environment) === 'test' ? $options['bankful_field_test_username'] : $options['bankful_field_live_username'];
	$req_password = strtolower($environment) === 'test' ? $options['bankful_field_test_password'] : $options['bankful_field_live_password'];
	if (!isset($site_id) || empty($site_id)) {
		$response = wp_remote_post(
			$request_site_id,
			array(
				'body' => array( 'req_username' => $req_username, 'req_password' => $req_password, 'site_url' => get_site_url(), 'site_title' => get_bloginfo('name') ),
				'headers'     => array(
					'Content-Type' => 'application/x-www-form-urlencoded',
				),
			)
		);
		$body = wp_remote_retrieve_body($response);
		$json_response = json_decode($body);
		$data_response = $json_response->data;
		if (isset($data_response) && !empty($data_response)) {
			write_logs_bf($data_response, 'info');
			if ($data_response->site_id) {
				write_logs_bf($data_response->site_id, 'info');
				add_option('bankful_options_site_id', $data_response->site_id);
			}
		}
		write_logs_bf($body, 'info');
	}
	
	//////////////////
	write_logs_bf('Start Log Migration--------------------------');
	$params = array(
		'type' => array( 'simple', 'subscription' ),
		'post_status' => array( 'publish' ),
		'recurring_product' => 'yes',
	);

	write_logs_bf($params);

	$updatedCount = 0;
	$products = wc_get_products($params);

	foreach ($products as $product) {
		// $product = wc_get_product($product->get_id());
		write_logs_bf($product->get_ID());
		write_logs_bf($product->get_title());
		write_logs_bf($product->get_type());
		
		$fetchObj = new stdClass();
		$fetchObj->req_username = $req_username;
		$fetchObj->req_password = $req_password;
		$fetchObj->site_id = $site_id;
		$fetchObj->product_id = $product->get_id();
		$fetchObj->cart_id = 10006;
		write_logs_bf(print_r($fetchObj, true));
		$ch = curl_init();
		
		curl_setopt($ch, CURLOPT_URL, $api_endpoint . 'handle-webhook/cart/woocommerce/verify_rules');
		curl_setopt($ch, CURLOPT_POST, 1);
		curl_setopt($ch, CURLOPT_HTTPHEADER, array( 'Content-Type:application/json' ));
		curl_setopt($ch, CURLOPT_POSTFIELDS, json_encode( $fetchObj ));
		
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
		
		$server_output = curl_exec($ch);

		$server_output = json_decode($server_output);

		// wp_set_object_terms( get_the_ID(), 'subscription', 'product_type' );
		$updatedCount = ++$updatedCount;
		write_logs_bf(print_r($server_output, true));

		wp_remove_object_terms( 65, 'simple', 'product_type', true );
		wp_set_object_terms( 65, 'subscription', 'product_type', true );
		
		$product->update_meta_data('_subscription_length', $server_output->expire_after);
		$product->update_meta_data('_subscription_period_interval', $server_output->frequency);
		$product->update_meta_data('_subscription_period', $server_output->rebill);
		
		$product->save();

		wp_remove_object_terms( $product->get_ID(), 'simple', 'product_type', true );
		wp_set_object_terms( $product->get_ID(), 'subscription', 'product_type', true );
		
		curl_close ($ch);
	}

	write_logs_bf($updatedCount);
	wp_send_json( array( 'updatedCount' => $updatedCount ), 200, null );
}

/**
 * Bankful field callback function.
 *
 * @param array $args
 */
function bankful_field_test_cb( $args ) {
	// Get the value of the setting we've registered with register_setting()
	$options = get_option('bankful_options');
	?>
	<input id="<?php echo esc_attr($args['label_for']); ?>" name="bankful_options[<?php echo esc_attr($args['label_for']); ?>]" value="<?php echo isset($options[$args['label_for']]) ? esc_attr($options[$args['label_for']]) : esc_attr($args['default']); ?>" type="<?php echo esc_attr($args['type']); ?>" class="<?php echo isset($args['class']) ? esc_attr($args['class']) : ''; ?> " <?php echo ( isset($args['info']) && !empty($args['info']) ) ? 'readonly' : ''; ?> />
	<?php
	if (isset($args['info']) && !empty($args['info']) && ( !empty($options['bankful_field_test_validate']) || !empty($options['bankful_field_live_validate']) )) {

		$accounts_array = array();
		( isset($options['bankful_field_test_validated_account_cc']) && 'yes' == $options['bankful_field_test_validated_account_cc'] ) ?
			array_push($accounts_array, 'Credit Card') : '';
		( isset($options['bankful_field_test_validated_account_hg']) && 'yes' == $options['bankful_field_test_validated_account_hg'] ) ?
			array_push($accounts_array, 'Hosted Page') : '';

		$accounts_array_live = array();
		( isset($options['bankful_field_live_validated_account_cc']) && 'yes' == $options['bankful_field_live_validated_account_cc'] ) ?
			array_push($accounts_array_live, 'Credit Card') : '';
		( isset($options['bankful_field_live_validated_account_hg']) && 'yes' == $options['bankful_field_live_validated_account_hg'] ) ?
			array_push($accounts_array_live, 'Hosted Page') : '';

		echo ( 'TEST' == $args['info'] ) ? ( ( isset($options['bankful_field_test_subscription_status']) && 'yes' == $options['bankful_field_test_subscription_status'] ) ?
			'<span class="dashicons dashicons-update-alt" title="Test Subscription Enabled"></span>' : '' ) : '';
		echo ( 'LIVE' == $args['info'] ) ? ( ( isset($options['bankful_field_live_subscription_status']) && 'yes' == $options['bankful_field_live_subscription_status'] ) ?
			'<span class="dashicons dashicons-update-alt" title="Live Subscription Enabled"></span>' : '' ) : '';
		?>
		<span class="dashicons dashicons-yes-alt" style="color: green;" title="Activated"></span>
		<span class="block"><?php echo esc_html($args['info']); ?> Environment credentials activated with accounts:
			<?php echo esc_html(implode(', ', ( 'TEST' == $args['info'] ) ? $accounts_array : $accounts_array_live)); ?>
			(<a id="<?php echo esc_attr($args['info']); ?>_edit" href="">Update</a>)
		</span>
	<?php
	}
	?>
	<p class="description <?php echo esc_attr($args['label_for']); ?>">
		<?php esc_html_e($args['desc'], 'bankful'); ?>
	</p>
<?php
}

/**
 * Developers section callback function.
 *
 * @param array $args  The settings array, defining title, id, callback.
 */
function bankful_section_settings_callback( $args ) {
	$site_id = get_option( 'bankful_options_site_id' );
	?>
	<p id="<?php echo esc_attr($args['id']); ?>"><?php esc_html_e('Here you can define TEST and LIVE Credentials for your plugin. ', 'bankful'); ?><?php echo 'Site ID: ' . esc_html($site_id) ; ?></p>
<?php
}

function bankful_activate() {
	global $woocommerce;
	global $wp_version;
	global $LIVE_API_SRV;

	$request_url = $LIVE_API_SRV . 'api/woocommerce/record-plugin-activity';

	$params = array(
		'woo_version' => isset($woocommerce->version) ? $woocommerce->version : '',
		'request_type' => 'A',
		'bf_site_url' => get_site_url(),
		'wp_version' => $wp_version,
		'php_version' => phpversion(),
		'bf_plugin_source' => 'woo_market',
		'site_admin_email' =>  get_bloginfo('admin_email'),
		'hash_salt' => '6a631f095832598284f22b5be8c0cc2e',
	);
	$response = wp_remote_post(
		$request_url,
		array(
			'body' => http_build_query($params),
			'headers'     => array(
				'Content-Type' => 'application/x-www-form-urlencoded',
			),
		)
	);

	$body = wp_remote_retrieve_body($response);
}
register_activation_hook(__FILE__, 'bankful_activate');

function bankful_deactivate() {
	global $woocommerce;
	global $wp_version;
	global $LIVE_API_SRV;

	$request_url = $LIVE_API_SRV . 'api/woocommerce/record-plugin-activity';

	$params = array(
		'woo_version' => $woocommerce->version,
		'request_type' => 'D',
		'bf_site_url' => get_site_url(),
		'wp_version' => $wp_version,
		'php_version' => phpversion(),
		'bf_plugin_source' => 'woo_market',
		'site_admin_email' =>  get_bloginfo('admin_email'),
		'hash_salt' => '6a631f095832598284f22b5be8c0cc2e',
	);
	$response = wp_remote_post(
		$request_url,
		array(
			'body' => http_build_query($params),
			'headers'     => array(
				'Content-Type' => 'application/x-www-form-urlencoded',
			),
		)
	);

	$body = wp_remote_retrieve_body($response);
}
register_deactivation_hook( __FILE__, 'bankful_deactivate' );


// API endpoint for syncing Orders
add_action('rest_api_init', function () {
	register_rest_route( 'bankful/v1', 'recurring-order/', array(
				'methods'  => 'POST',
				'callback' => 'create_recurring_order_bankful',
				'permission_callback' => 'check_permission_rest_create_order',
	));
});

function check_permission_rest_create_order( $request ) {
	// Check if the parameter has the expected value
	if (isset($request['req_username']) && isset($request['req_password'])) {
		return true; // Permission granted
	} else {
		return false; // Permission denied
	}
}

function create_recurring_order_bankful( $request ) {

	$status = $request['status'];

	if ('pending' == $status) {
		$args = array(
			'req_username' => $request['req_username'],
			'req_password' => $request['req_password'],
			'product' => $request['product'],
			'customer_email' => $request['customer_email'],
			'billing_address' => $request['billing_address'],
			'shipping_address' => $request['shipping_address'],
		);
		$pending = true;
	} else {
		$args = array(
			'req_username' => $request['req_username'],
			'req_password' => $request['req_password'],
			'transaction_id' => $request['transaction_id'],
			'order_id' => $request['order_id'],
		);
		$complete = true;
	}

	$options = get_option( 'bankful_options' );
	$environment = $options['bankful_field_environment'];

	$req_username = strtolower($environment) === 'test' ? $options['bankful_field_test_username'] : $options['bankful_field_live_username'];
	$req_password = strtolower($environment) === 'test' ? $options['bankful_field_test_password'] : $options['bankful_field_live_password'];

	$payment_gateway_id = get_current_active_gateway_id();

	$payment_gateway = WC()->payment_gateways->payment_gateways()[$payment_gateway_id];
	$method_title = $payment_gateway->method_title;

	if (strcmp($req_username, $args['req_username']) !== 0) {
		return new WP_Error( 'bankful_merchant_error', 'incorrect credentials', array( 'status' => 404 ) );
	}

	if (strcmp($req_password, $args['req_password']) !== 0) {
		return new WP_Error( 'bankful_merchant_error', 'incorrect credentials', array( 'status' => 404 ) );
	}

	if ($pending) {

		$order_email = $args['customer_email'];

		$billing_address = array(
			'first_name' => $args['billing_address']['first_name'],           
			'last_name'  => $args['billing_address']['last_name'],     
			'email'      => $order_email,
			'phone'      => $args['billing_address']['phone'],   
			'address_1'  => $args['billing_address']['address1'],     
			'address_2'  => $args['billing_address']['address2'],             
			'city'       => $args['billing_address']['city'],     
			'state'      => $args['billing_address']['state'],             
			'postcode'   => $args['billing_address']['zip'],            
			'country'    => $args['billing_address']['phone'],                
		);

		$shipping_address = array(
			'first_name' => $args['shipping_address']['first_name'],
			'last_name'  => $args['shipping_address']['last_name'],
			'email'      => $order_email,
			'phone'      => $args['shipping_address']['phone'],
			'address_1'  => $args['shipping_address']['address1'],
			'address_2'  => $args['shipping_address']['address2'],
			'city'       => $args['shipping_address']['city'],
			'state'      => $args['shipping_address']['state'],
			'postcode'   => $args['shipping_address']['zip'],
			'country'    => $args['shipping_address']['phone'],
		);

		$user = get_user_by('email', $order_email);

		if ( !$user ) {
			$default_password = wp_generate_password();
			$user = wp_create_user( $order_email, $default_password, $order_email );
		}
	
		$order = wc_create_order(array( 'customer_id' => $user->id ));
	
		$order->add_product( wc_get_product($args['product']), 1); // This is an existing SIMPLE product
		$order->set_address( $billing_address, 'billing' );
		$order->set_address( $shipping_address, 'shipping' );

		$order->calculate_totals();

		$note = __('This is Recurring Order, Created By Bankful Recurring Service and waiting for Payment, If Product is Physical do not ship the Product until successful payment.');
	}

	if ($complete) {
		$order = wc_get_order( $args['order_id'] );

		$order->set_payment_method( $method_title );
		$order->set_payment_method_title( $method_title );
		$order->set_transaction_id( $args['transaction_id'] );
		$order->set_date_paid( time() );

		$note = __('This is Recurring Order, Completed By Bankful Recurring Service after Successful Payment, If Product is Physical you can now ship the Product.');
	}

	// Add the note
	$order->add_order_note( $note );

	update_post_meta($order->id, 'bfp_recurring', true);

	$order->update_status($status, 'Recurring order', true);  

	return $order->id;
}

// API endpoint for cancelling subscription
add_action('rest_api_init', function () {
	register_rest_route( 'bankful/v1', 'stop-recurring-order/', array(
		'methods'  => 'POST',
		'callback' => 'stop_bankful_subscription',
		'permission_callback' => 'check_permission_rest_stop_order',
	));
});

function check_permission_rest_stop_order( $request ) {
	// Check if the parameter has the expected value
	if (isset($request['recurringSubscriptionId'])) {
		return true; // Permission granted
	} else {
		return false; // Permission denied
	}
}

function stop_bankful_subscription( $request ) {

	$options = get_option( 'bankful_options' );
	$environment = $options['bankful_field_environment'];
	
	$req_username = strtolower($environment) === 'test' ? $options['bankful_field_test_username'] : $options['bankful_field_live_username'];
	$req_password = strtolower($environment) === 'test' ? $options['bankful_field_test_password'] : $options['bankful_field_live_password'];

	if (!strcmp($req_username, $request['req_username']) === 0 || !strcmp($req_password, $request['req_password']) === 0) {
		return 'Merchant Not found';
	}

	$subsOrderID = $request['recurringSubscriptionId'];
	write_logs_bf($request, 'info');

	// $order = wc_get_order( $subsOrderID ); 
	if (!class_exists('WC_Subscription')) {
		return 'Bankful Subscription is not enabled on this store.';
	}

	$subscription = new WC_Subscription( $subsOrderID );

	if ( $subscription->get_id() ) {
		write_logs_bf(print_r($subscription, true), 'info');
		$subscription->update_status( 'on-hold' );
		// $subscriptions = wcs_get_subscriptions_for_order( $subsOrderID );

		// foreach ( $subscriptions as $subscription_id => $subscription ) {
		//  // Change the status of the WC_Subscription object
		// }

	} else {
		return 'Bankful Subscription ID not found.';
	}

	return $subsOrderID;
}

function get_current_active_gateway_id() {
	$options = get_option( 'bankful_options' );
	$active_gateway_id = 'bankful_credit_card';
	if (isset($options['bankful_field_environment']) && 'Test' == $options['bankful_field_environment']) {

		( isset($options['bankful_field_test_validated_account_cc']) && 'yes' == $options['bankful_field_test_validated_account_cc'] ) ?
		$active_gateway_id = 'bankful_credit_card' : '';
		( isset($options['bankful_field_test_validated_account_hg']) && 'yes' == $options['bankful_field_test_validated_account_hg'] ) ?
		$active_gateway_id = 'bankful_hosted_gateway' : '';

	} else if (isset($options['bankful_field_environment']) && 'Live' == $options['bankful_field_environment']) {

		( isset($options['bankful_field_live_validated_account_cc']) && 'yes' == $options['bankful_field_live_validated_account_cc'] ) ?
		$active_gateway_id = 'bankful_credit_card' : '';
		( isset($options['bankful_field_live_validated_account_hg']) && 'yes' == $options['bankful_field_live_validated_account_hg'] ) ?
		$active_gateway_id = 'bankful_hosted_gateway' : '';

	}

	return $active_gateway_id;
}

add_action( 'woocommerce_subscription_renewal_payment_complete', 'subscription_renewal_payment_complete', 11, 2 );
function subscription_renewal_payment_complete( $subscription, $order ) {
	global $SANDBOX_API_SRV;
	global $LIVE_API_SRV;

	$options = get_option( 'bankful_options' );
	$site_id = get_option( 'bankful_options_site_id' );
	$environment = $options['bankful_field_environment'];
	
	$api_endpoint = strtolower($environment) === 'test' ? $SANDBOX_API_SRV . 'api/' : $LIVE_API_SRV . 'api/';

	$request_url = $api_endpoint . 'woocommerce/sync-subscription';
	$req_username = strtolower($environment) === 'test' ? $options['bankful_field_test_username'] : $options['bankful_field_live_username'];
	$req_password = strtolower($environment) === 'test' ? $options['bankful_field_test_password'] : $options['bankful_field_live_password'];

	$subscription_id    = $subscription->get_id();
	$next_payment       = $subscription->get_date('next_payment', 'site');

	$fetchObj = new stdClass(); 

	$fetchObj->req_username = $req_username;
	$fetchObj->req_password = $req_password;
	$fetchObj->site_id = $site_id;
	$fetchObj->next_payment_date = $next_payment;
	$fetchObj->subscription_id = $subscription_id;

	$ch = curl_init();
					
	curl_setopt($ch, CURLOPT_URL, $request_url);
	curl_setopt($ch, CURLOPT_POST, 1);
	curl_setopt($ch, CURLOPT_HTTPHEADER, array( 'Content-Type:application/json' ));
	curl_setopt($ch, CURLOPT_POSTFIELDS, json_encode( $fetchObj ));
	
	curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
	
	$server_output = curl_exec($ch);

	$body_pre = json_decode($server_output);
	
	curl_close ($ch);
}

add_action( 'woocommerce_subscription_status_updated', 'update_bankful_on_status_change', 11, 3);
function update_bankful_on_status_change( $subscription, $new_status, $old_status ) {
	global $SANDBOX_API_SRV;
	global $LIVE_API_SRV;

	$options = get_option( 'bankful_options' );
	$site_id = get_option( 'bankful_options_site_id' );
	$environment = $options['bankful_field_environment'];
	
	$api_endpoint = strtolower($environment) === 'test' ? $SANDBOX_API_SRV . 'api/' : $LIVE_API_SRV . 'api/';

	$request_url = $api_endpoint . 'woocommerce/stop-subscription';
	$req_username = strtolower($environment) === 'test' ? $options['bankful_field_test_username'] : $options['bankful_field_live_username'];
	$req_password = strtolower($environment) === 'test' ? $options['bankful_field_test_password'] : $options['bankful_field_live_password'];
	write_logs_bf('update_bankful_on_status_change STOP Recurring ID: ' . $subscription->get_id() . '. new status: ' . $new_status . '. Old Status: ' . $old_status, 'info');
	if (in_array($new_status, array( 'cancel', 'cancelled', 'expired' ))) {
		$fetchObj = new stdClass(); 
	
		$fetchObj->req_username = $req_username;
		$fetchObj->req_password = $req_password;
		$fetchObj->site_id = $site_id;
		$fetchObj->subscription_id = $subscription->get_id();
	
		$ch = curl_init();
						
		curl_setopt($ch, CURLOPT_URL, $request_url);
		curl_setopt($ch, CURLOPT_POST, 1);
		curl_setopt($ch, CURLOPT_HTTPHEADER, array( 'Content-Type:application/json' ));
		curl_setopt($ch, CURLOPT_POSTFIELDS, json_encode( $fetchObj ));
		
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
		
		$server_output = curl_exec($ch);
	
		$body_pre = json_decode($server_output);
		
		curl_close ($ch);
	}
}

/**
 * Handle a custom 'customvar' query var to get products with the 'customvar' meta.
 *
 * @param array $query - Args for WP_Query.
 * @param array $query_vars - Query vars from WC_Product_Query.
 * @return array modified $query
 */
function handle_custom_query_var( $query, $query_vars ) {
	if ( ! empty( $query_vars['recurring_product'] ) ) {
		$query['meta_query'][] = array(
			'key' => 'recurring_product',
			'value' => esc_attr( $query_vars['recurring_product'] ),
		);
	}

	return $query;
}
add_filter( 'woocommerce_product_data_store_cpt_get_products_query', 'handle_custom_query_var', 10, 2 );

// add_action('woocommerce_order_status_changed', 'empty_cart_on_cancel', 10, 3);
// function empty_cart_on_cancel($order_id, $old_status, $new_status) {
// write_logs_bf('woocommerce_cancel_unpaid_order CALLED');
// write_logs_bf($new_status);
// write_logs_bf($old_status);
// WC()->cart->empty_cart();
// die();
// }

//Suspension and Reactivation
add_filter('wcs_view_subscription_actions', 'add_suspend_subscription_button', 10, 3);

function add_suspend_subscription_button( $actions, $subscription, $user_id ) {

	if ( ! $subscription->can_be_updated_to( 'on-hold' ) ) {
		return $actions;
	}

	if ( ! user_can( $user_id, 'edit_shop_subscription_status', $subscription->get_id() ) ) {
		return $actions;
	}

	if ( current_user_can( 'manage_woocommerce' ) || wcs_can_user_put_subscription_on_hold( $subscription, $user_id ) ) {
		if ($subscription->has_status(array( 'active', 'on-hold' ))) {
			$actions['suspend'] = array(
				'url'  => wp_nonce_url(add_query_arg('suspend_subscription', $subscription->get_id()), 'suspend_subscription'),
				'name' => __('Suspend', 'woocommerce-subscriptions'),
			);
		}
	}
	return $actions;
}

add_action('init', 'handle_suspend_subscription_action');

function handle_suspend_subscription_action() {
	if (isset($_GET['suspend_subscription']) && isset($_GET['_wpnonce'])) {
		if (!wp_verify_nonce(sanitize_key($_GET['_wpnonce']), 'suspend_subscription')) {
			return;
		}

		$subscription_id = absint($_GET['suspend_subscription']);
		$subscription = wcs_get_subscription($subscription_id);

		if ($subscription && $subscription->has_status(array( 'active', 'on-hold' ))) {
			$subscription->update_status('on-hold');
			wc_add_notice(__('Subscription suspended successfully.', 'woocommerce-subscriptions'), 'success');
			wp_safe_redirect(wc_get_account_endpoint_url('subscriptions'));
			exit();
		}
	}
}

add_filter('woocommerce_subscription_actions', 'add_resume_subscription_button', 10, 2);

function add_resume_subscription_button( $actions, $subscription ) {
	if ($subscription->has_status('on-hold')) {
		$actions['resume'] = array(
			'url'  => wp_nonce_url(add_query_arg('resume_subscription', $subscription->get_id()), 'resume_subscription'),
			'name' => __('Resume', 'woocommerce-subscriptions'),
		);
	}
	return $actions;
}

add_action('init', 'handle_resume_subscription_action');

function handle_resume_subscription_action() {
	if (isset($_GET['resume_subscription']) && isset($_GET['_wpnonce'])) {
		if (!wp_verify_nonce(sanitize_key($_GET['_wpnonce']), 'resume_subscription')) {
			return;
		}

		$subscription_id = absint($_GET['resume_subscription']);
		$subscription = wcs_get_subscription($subscription_id);

		if ($subscription && $subscription->has_status('on-hold')) {
			$subscription->update_status('active');
			wc_add_notice(__('Subscription resumed successfully.', 'woocommerce-subscriptions'), 'success');
			wp_safe_redirect(wc_get_account_endpoint_url('subscriptions'));
			exit();
		}
	}
}

// add_filter('wcs_view_subscription_actions', 'WC_Bankful_WS_Gateway::add_renew_now_button', 10, 2 );
// add_action('init', 'WC_Bankful_WS_Gateway::handle_renew_subscription_action');

// add_filter('wcs_view_subscription_actions', array(WC_Bankful_WS_Gateway::class, 'add_renew_now_button'), 10, 2 );
// add_action('init', array(WC_Bankful_WS_Gateway::class, 'handle_renew_subscription_action'));
add_filter('wcs_view_subscription_actions', 'add_renew_now_button', 10, 2);

function add_renew_now_button( $actions, $subscription ) {
	if ($subscription->has_status(array( 'active', 'on-hold' ))) {
		$actions['renew_now'] = array(
			'url'  => wp_nonce_url(add_query_arg('renew_subscription', $subscription->get_id()), 'renew_subscription'),
			'name' => __('Renew Now', 'woocommerce-subscriptions'),
		);
	}
	return $actions;
}

add_action('init', 'handle_renew_subscription_action');

function handle_renew_subscription_action() {
	if (isset($_GET['renew_subscription']) && isset($_GET['_wpnonce'])) {
		if (!wp_verify_nonce(sanitize_key($_GET['_wpnonce']), 'renew_subscription')) {
			return;
		}

		$subscription_id = absint($_GET['renew_subscription']);
		$subscription = wcs_get_subscription($subscription_id);

		if ($subscription && $subscription->has_status(array( 'active', 'on-hold' ))) {
			$order = wcs_create_renewal_order($subscription);
			$order_total = $order->get_total();
			if ($order) {
				if (class_exists('WC_Bankful_WS_Gateway')) {
					$class_obj = new WC_Bankful_WS_Gateway();
					$class_obj->process_subscription_payment($order_total, $order);
				} else if (class_exists('WC_Gateway_Bankful')) {
					$class_obj = new WC_Gateway_Bankful();
					$class_obj->process_subscription_payment($order_total, $order);
				}
				// $subscription->update_dates(array('last_payment' => current_time('mysql', 1)));
				// Remove any overdue status
				$subscription->update_status('active');
				wc_add_notice(__('Subscription renewed successfully.', 'woocommerce-subscriptions'), 'success');
				wp_safe_redirect(wc_get_account_endpoint_url('subscriptions'));
				exit();
			} else {
				wc_add_notice(__('Failed to renew the subscription.', 'woocommerce-subscriptions'), 'error');
			}
		}
	}
}

// if (!function_exists('wcs_create_renewal_order')) {
//     function wcs_create_renewal_order($subscription) {
//         $renewal_order = wc_create_order(array('customer_id' => $subscription->get_customer_id()));
		
//         foreach ($subscription->get_items() as $item_id => $item) {
//             $product = $item->get_product();
//             $renewal_order->add_product($product, $item->get_quantity(), array('totals' => $item->get_totals()));
//         }

//         foreach ($subscription->get_fees() as $fee_id => $fee) {
//             $renewal_order->add_fee($fee);
//         }

//         foreach ($subscription->get_taxes() as $tax_id => $tax) {
//             $renewal_order->add_tax($tax);
//         }

//         $renewal_order->set_address($subscription->get_address('billing'), 'billing');
//         $renewal_order->set_address($subscription->get_address('shipping'), 'shipping');

//         $renewal_order->calculate_totals();
//         $renewal_order->update_status('pending');

//         return $renewal_order;
//     }
// }


// /**
//  * Checks the cart to see if it contains an early subscription renewal.
//  *
//  * @return bool|array The cart item containing the early renewal, else false.
//  * @since  2.3.0
//  */
// function wcs_cart_contains_early_renewal() {

//  $cart_item = wcs_cart_contains_renewal();

//  if ( $cart_item && ! empty( $cart_item['subscription_renewal']['subscription_renewal_early'] ) ) {
//      return $cart_item;
//  }

//  return false;
// }

// /**
//  * Checks if a user can renew an active subscription early.
//  *
//  * @param int|WC_Subscription $subscription Post ID of a 'shop_subscription' post, or instance of a WC_Subscription object.
//  * @param int $user_id The ID of a user.
//  * @since 2.3.0
//  * @return bool Whether the user can renew a subscription early.
//  */
// function wcs_can_user_renew_early( $subscription, $user_id = 0 ) {
//  $subscription = wcs_get_subscription( $subscription );
//  $user_id      = ! empty( $user_id ) ? $user_id : get_current_user_id();
//  $reason       = '';

//  // Check for all the normal reasons a subscription can't be renewed early.
//  if ( ! $subscription ) {
//      $reason = 'not_a_subscription';
//  } elseif ( ! $subscription->has_status( array( 'active' ) ) ) {
//      $reason = 'subscription_not_active';
//  } elseif ( 0.0 === floatval( $subscription->get_total() ) ) {
//      $reason = 'subscription_zero_total';
//  } elseif ( $subscription->get_time( 'trial_end' ) > gmdate( 'U' ) ) {
//      $reason = 'subscription_still_in_free_trial';
//  } elseif ( ! $subscription->get_time( 'next_payment' ) ) {
//      $reason = 'subscription_no_next_payment';
//  } elseif ( ! $subscription->payment_method_supports( 'subscription_date_changes' ) ) {
//      $reason = 'payment_method_not_supported';
//  } elseif (
//      WC_Subscriptions_Synchroniser::subscription_contains_synced_product( $subscription ) &&
//      /**
//       * Determine whether a subscription with Synchronized products can be renewed early.
//       *
//       * @param bool            $can_renew_early Whether the subscription can be renewed early.
//       * @param WC_Subscription $subscription    The subscription to be renewed early.
//       */
//      ! boolval( apply_filters( 'wcs_allow_synced_product_early_renewal', false, $subscription ) )
//  ) {
//      $reason = 'subscription_contains_synced_product';
//  }

//  // Make sure all line items still exist.
//  foreach ( $subscription->get_items() as $line_item ) {
//      $product = wc_get_product( wcs_get_canonical_product_id( $line_item ) );

//      if ( false === $product ) {
//          $reason = 'line_item_no_longer_exists';
//          break;
//      }
//  }

//  // Non-empty $reason means we can't renew early.
//  $can_renew_early = empty( $reason );

//  /**
//   * Allow third-parties to filter whether the customer can renew a subscription early.
//   *
//   * @since 2.3.0
//   *
//   * @param bool            $can_renew_early Whether early renewal is permitted.
//   * @param WC_Subscription $subscription    The subscription being renewed early.
//   * @param int             $user_id         The user's ID.
//   * @param string          $reason          The reason why the subscription cannot be renewed early. Empty
//   *                                         string if the subscription can be renewed early.
//   */
//  return apply_filters( 'woocommerce_subscriptions_can_user_renew_early', $can_renew_early, $subscription, $user_id, $reason );
// }

// /**
//  * Returns a URL for early renewal of a subscription.
//  *
//  * @param  int|WC_Subscription $subscription WC_Subscription ID, or instance of a WC_Subscription object.
//  * @return string The early renewal URL.
//  * @since  2.3.0
//  */
// function wcs_get_early_renewal_url( $subscription ) {
//  $subscription_id = is_a( $subscription, 'WC_Subscription' ) ? $subscription->get_id() : absint( $subscription );

//  $url = add_query_arg( array(
//      'subscription_renewal_early' => $subscription_id,
//      'subscription_renewal'       => 'true',
//  ), get_permalink( wc_get_page_id( 'myaccount' ) ) );

//  /**
//   * Allow third-parties to filter the early renewal URL.
//   *
//   * @since 2.3.0
//   * @param string $url The early renewal URL.
//   * @param int    $subscription_id The ID of the subscription to renew to.
//   */
//  return apply_filters( 'woocommerce_subscriptions_get_early_renewal_url', $url, $subscription_id ); // nosemgrep: audit.php.wp.security.xss.query-arg -- False positive. $url is escaped in the template and escaping URLs should be done at the point of output or usage.
// }

// /**
//  * Update the subscription dates after processing an early renewal.
//  *
//  * @since 2.6.0
//  *
//  * @param WC_Subscription $subscription The subscription to update.
//  * @param WC_Order $early_renewal       The early renewal.
//  */
// function wcs_update_dates_after_early_renewal( $subscription, $early_renewal ) {
//  $dates_to_update = WCS_Early_Renewal_Manager::get_dates_to_update( $subscription );

//  if ( ! empty( $dates_to_update ) ) {
//      // translators: %s: order ID.
//      $order_number = sprintf( _x( '#%s', 'hash before order number', 'woocommerce-subscriptions' ), $early_renewal->get_order_number() );
//      $order_link   = sprintf( '<a href="%s">%s</a>', esc_url( wcs_get_edit_post_link( $early_renewal->get_id() ) ), $order_number );

//      try {
//          $subscription->update_dates( $dates_to_update );

//          // translators: placeholder contains a link to the order's edit screen.
//          $subscription->add_order_note( sprintf( __( 'Customer successfully renewed early with order %s.', 'woocommerce-subscriptions' ), $order_link ) );
//      } catch ( Exception $e ) {
//          // translators: placeholder contains a link to the order's edit screen.
//          $subscription->add_order_note( sprintf( __( 'Failed to update subscription dates after customer renewed early with order %s.', 'woocommerce-subscriptions' ), $order_link ) );
//      }
//  }
// }




add_action('before_woocommerce_init', function () {
	if (class_exists(\Automattic\WooCommerce\Utilities\FeaturesUtil::class)) {
		\Automattic\WooCommerce\Utilities\FeaturesUtil::declare_compatibility('custom_order_tables', __FILE__, true);
		\Automattic\WooCommerce\Utilities\FeaturesUtil::declare_compatibility('cart_checkout_blocks', __FILE__, true);
	}
});
